/* 
 * mm.c - Micro Memory(tm) PCI memory board block device driver - v1.00
 *
 * (C) 2001 San Mehat <nettwerk@valinux.com>
 * (C) 2001 Johannes Erdfelt <jerdfelt@valinux.com>
 *
 * This driver provides a standard block device interface for Micro Memory(tm)
 * PCI based RAM boards..
 */

#include <linux/config.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/hdreg.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/mman.h>
#include <linux/slab.h>
#include <linux/ioctl.h>
#include <linux/fd.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/smp_lock.h>
#include <linux/timer.h>
#include <linux/spinlock.h>
#include <linux/pci.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/byteorder.h>
#include <linux/delay.h>
#include "mm.h"

static unsigned char get_window(int card);
static void set_window(int card, unsigned char num);
static int do_dma(int card,unsigned long host,unsigned long local,unsigned int size, int dir);
static void mm_interrupt(int irq, void *__card, struct pt_regs *regs);
static int window_test(int card);
static int set_removeled(int card, unsigned char state);
static void dump_regs(int card);
static void dump_card(int card, int nbytes);
static void dump_dmastat(int card, unsigned int dmastat);
static int dword_test(int card);
static void mm_int_timer(unsigned long data);

#define MAJOR_NR JOYSTICK_MAJOR
#include <linux/blk.h>
#include <linux/blkpg.h>

static unsigned long  rd_len[MM_MAXCARDS];
static int            rd_hardsect[MM_MAXCARDS];
static int            rd_blocksizes[MM_MAXCARDS];
static int            rd_kbsize[MM_MAXCARDS];
static int            num_cards = 0;

static int device_id[] = { 0x5415};

static DECLARE_MUTEX(mm_sem);

struct cardinfo {
  int            card_number;
  struct pci_dev *dev;
  unsigned long  csr_base;
  unsigned long  csr_remap;
  unsigned long  csr_len;
#ifdef CONFIG_MM_MAP_MEMORY
  unsigned long  mem_base;
  unsigned long  mem_remap;
#endif
  unsigned long  mem_len;

  unsigned int   rd_size;  // size in kbytes
  unsigned int   win_size; // PCI window size
  int            irq;
#ifdef DECLARE_WAITQUEUE
  wait_queue_head_t wait;
#else
  struct wait_queue wait;
#endif
  int            error;
  unsigned char  *dma_buffer;
  struct timer_list irq_timer;
};

static struct cardinfo cards[MM_MAXCARDS];

int rd_blocksize = 4096;

static int mm_make_request(request_queue_t *q, int rw, struct buffer_head *sbh)
{
  unsigned int  minor;
  unsigned long offset, len;
  void          *host;
  void          *hosthigh = NULL;

  minor = MINOR(sbh->b_rdev);
  if (minor >= num_cards)
    goto fail;

  offset = sbh->b_rsector << 9;
  len = sbh->b_size;
  if ((offset + len) > rd_len[minor]) {
    printk("MM%d: Block request past end of device (offset 0x%lx (%ld))\n",
            minor,
            offset,
            offset);
    goto fail;
  }

  if (rw == READA)
    rw = READ;

  if ((rw != READ) && (rw != WRITE)) {
    printk("MM: bad command: %d\n",rw);
    goto fail;
  }

#if 0
  printk("MM%d: BLK_RQ o = %ld, l = %ld op = %s max = %ld\n",
          minor,
          offset,
          len,
          (rw == READ ? "R" : "W"),
          rd_len[minor]);
#endif
  hosthigh = (void *) bh_kmap(sbh);

  if (rw == READ) {
    host  = (void *) cards[minor].dma_buffer;
    if (len > PAGE_SIZE)
      goto fail;
    if (do_dma(minor,
           virt_to_bus(host),
           offset,
           len,
           DMA_WRITE_TO_HOST)!=0)
      goto fail;
    memcpy(hosthigh, host, len);
  } else {
    host  = (void *) cards[minor].dma_buffer;
    if (len > PAGE_SIZE)
      goto fail;
    memcpy(host, hosthigh, len);
    if (do_dma(minor,
               virt_to_bus(host),
               offset,
               len,
               DMA_READ_FROM_HOST) != 0)
      goto fail;
  }

  bh_kunmap(sbh);
  sbh->b_end_io(sbh, 1);
  return 0;

fail:
  if (hosthigh)
    bh_kunmap(sbh);

  sbh->b_end_io(sbh, 0);
  return 0;
}

static int do_copy_to_card(int c, unsigned long host, unsigned long local, unsigned int n)
{
  return 0;
}

static int do_copy_to_host(int c, unsigned long host, unsigned long local, unsigned int n)
{
  return 0;
}

static int do_dma(int card, unsigned long host,unsigned long local,unsigned int size, int dir)
{
  volatile struct dma_ctrl_stat *dma_reg;
  unsigned int                  *dma_count, *dma_host, *dma_local, *dma_chain, *dma_sem;
  unsigned int                  *dma_ctrl;
  int                           retries = 5;

  down(&mm_sem);
  while (retries) {
#if 0
    printk("MM%d: DMA op %d, host 0x%lx, local 0x%lx, size 0x%ld\n",card,dir,host,local,size);
#endif
    dma_ctrl = (unsigned int *) ((unsigned long) 
                cards[card].csr_remap+DMA_STATUS_CTRL);

    dma_sem = (unsigned int *) ((unsigned long) 
                cards[card].csr_remap+DMA_SEMAPHORE_ADDR);

    dma_count = (unsigned int *) ((unsigned long) 
                cards[card].csr_remap+DMA_TRANSFER_SIZE);
    dma_host = (unsigned int *) ((unsigned long) 
                cards[card].csr_remap+DMA_PCI_ADDR);
    dma_local = (unsigned int *) ((unsigned long) 
                cards[card].csr_remap+DMA_LOCAL_ADDR);
    dma_chain = (unsigned int *) ((unsigned long) 
                cards[card].csr_remap+DMA_DESCRIPTOR_ADDR);

    dma_reg = (struct dma_ctrl_stat *) ((unsigned long) 
              cards[card].csr_remap + DMA_STATUS_CTRL);
  
    *dma_host = (unsigned int) host;
    *(dma_host+1) = (unsigned int) 0x00;

    *dma_local = (unsigned int) local;
    *(dma_local+1) = (unsigned int) 0x00;

    *dma_count = size;
    *(dma_count+1) = (unsigned int) 0x00;

    *dma_chain = 0;
    *(dma_chain+1) = (unsigned int) 0x00;

    *dma_sem = 0;
    *(dma_sem+1) = (unsigned int) 0x00;

//  dump_regs(card);

    cards[card].irq_timer.expires = jiffies + IRQ_TIMEOUT;
    add_timer(&cards[card].irq_timer);
#ifdef LED_DIAG
    set_removeled(card, 0x01);
#endif
    if (dir == DMA_READ_FROM_HOST)
      *dma_ctrl = 0x63;
    else if (dir == DMA_WRITE_TO_HOST)
      *dma_ctrl = 0x61;
    else {
      printk("MM%d: Bad DMA direction specified\n",card);
      return -1;
    }

    sleep_on(&cards[card].wait);
    del_timer(&cards[card].irq_timer);
#ifdef LED_DIAG
    set_removeled(card, 0x00);
#endif
    if (cards[card].error != 0) {
#ifdef CONFIG_MM_INT_DEBUG
      printk("MM%d: Retrying DMA due to error %d (retries left %d)\n",
             card,
             cards[card].error,
             retries);
#endif
      retries--;
    } else
      break;
  }
  up(&mm_sem);

  if (cards[card].error != 0) {
    cards[card].error = 0;
    return 1;
  }

  return 0;
}

static int mm_ioctl(struct inode  *i,
                    struct file   *f,
                    unsigned int  cmd,
                    unsigned long arg)
{
  unsigned int minor;

  if (!i || !i->i_rdev)
    return -EINVAL;

  minor = MINOR(i->i_rdev);

  switch(cmd) {
  case BLKFLSBUF:
    break;
  case BLKGETSIZE:
    if (!arg)
      return -EINVAL;
    return put_user(rd_kbsize[minor] << 1, (long *) arg);
  case BLKROSET:
  case BLKROGET:
  case BLKSSZGET:
    return blk_ioctl(i->i_rdev, cmd, arg);
  default:
    return -EINVAL;
  }

  return 0;
}

static int mm_open(struct inode *i, struct file *filp)
{
  if (DEVICE_NR(i->i_rdev) >= num_cards)
    return -ENXIO;

  MOD_INC_USE_COUNT;
  return 0;
}

static int mm_do_release(struct inode *i, struct file *filp)
{
  MOD_DEC_USE_COUNT;
  return 0;
}

static struct block_device_operations fd_fops = {
  open:      mm_open,
  release:   mm_do_release,
  ioctl:     mm_ioctl,
};

int __init mm_init(void)
{
  struct pci_dev *dev = NULL;
  int            i, count = 0, idx;

  printk("Micro Memory(tm) PCI RAM driver\n");
  if (!pci_present()) {
    printk("No PCI found\n");
    return -ENXIO;
  }

  for (idx = 0; idx < (sizeof(device_id) / sizeof(int)); idx++) {
    dev = NULL;
    for (;;) {
      if (!(dev = pci_find_device(0x1332, device_id[idx], dev)))
        break;
      if (count >= MM_MAXCARDS)
        break;
      if (pci_enable_device(dev))
        break;
      if (!register_pci(count,dev))
        count++;
    }
  }

  num_cards = count;
  if (register_blkdev(MAJOR_NR, "mm", &fd_fops)) {
    printk("MM: Could not get major %d\n",MAJOR_NR);
    return -EIO;
  }

  blk_queue_make_request(BLK_DEFAULT_QUEUE(MAJOR_NR), &mm_make_request);
  for (i = 0; i < num_cards; i++) {
    rd_len[i] = cards[i].rd_size << 10;
    rd_hardsect[i] = rd_blocksize;
    rd_blocksizes[i] = rd_blocksize;
    rd_kbsize[i] = cards[i].rd_size;
  }

  hardsect_size[MAJOR_NR] = rd_hardsect;
  blksize_size[MAJOR_NR] = rd_blocksizes;
  blk_size[MAJOR_NR] = rd_kbsize;

  return 0;
}

static void __exit mm_cleanup(void)
{
  int i;

  unregister_blkdev(MAJOR_NR, "mm");
  hardsect_size[MAJOR_NR]= NULL;
  blksize_size[MAJOR_NR] = NULL;
  blk_size[MAJOR_NR] = NULL;

  for (i = 0; i < num_cards; i++) {
    iounmap((void *) cards[i].csr_remap);
    release_mem_region(cards[i].csr_base, cards[i].csr_len);
#ifdef CONFIG_MM_MAP_MEMORY
    iounmap((void *) cards[i].mem_remap);
    release_mem_region(cards[i].mem_base, cards[i].mem_len);
#endif
    free_irq(cards[i].irq, &cards[i]);
    free_page((unsigned long) cards[i].dma_buffer);
  }
}

int __init register_pci(int i, struct pci_dev *dev)
{
  char           *str;
  unsigned char  *mem_present;
  unsigned char  *batt_status;
  unsigned char  *edc_ctrl;
  unsigned int   saved_bar, data;
  unsigned short cfg_word;

  if (!dev || pci_enable_device(dev))
    return 0;
 
  switch(dev->device) {
  case PCI_DEVICE_ID_5415CN:
    str = "PCI Mem Module (Battery Backup)";
    break;
  default:
    str = "(unknown/unsupported)";
    break;
  }

  cards[i].csr_base = pci_resource_start(dev, 0);
  cards[i].csr_len  = pci_resource_len(dev, 0);
#ifdef CONFIG_MM_MAP_MEMORY
  cards[i].mem_base = pci_resource_start(dev, 1);
#endif
  cards[i].mem_len  = pci_resource_len(dev, 1);

  printk("Micro Memory(tm) controller #%d found at %02x:%02x (%s)\n",
         i,
         dev->bus->number,
         dev->devfn,
         str);

  cards[i].dev = dev;
  cards[i].card_number = i;
  if (!request_mem_region(cards[i].csr_base, cards[i].csr_len, "Micro Memory")) {
    printk("MM%d: Unable to request memory region\n",i);
    return(-ENOMEM);
  }

  if (!(cards[i].csr_remap= (unsigned long) ioremap_nocache(cards[i].csr_base,cards[i].csr_len))) {
    release_mem_region(cards[i].csr_base, cards[i].csr_len);
    printk("MM%d: Unable to remap memory region\n",i);
    return(-ENOMEM);
  }

  printk("MM%d: CSR 0x%8lx -> 0x%8lx (0x%lx)\n",i,cards[i].csr_base, cards[i].csr_remap,cards[i].csr_len);

#ifdef CONFIG_MM_MAP_MEMORY
  if (!request_mem_region(cards[i].mem_base, cards[i].mem_len, "Micro Memory")) {
    iounmap((void *) cards[i].csr_base);
    release_mem_region(cards[i].csr_base, cards[i].csr_len);
    printk("MM%d: Unable to request memory region\n",i);
    return(-ENOMEM);
  }

  if (!(cards[i].mem_remap=(unsigned long)ioremap(cards[i].mem_base,cards[i].mem_len))) {
    iounmap((void *) cards[i].csr_remap);
    release_mem_region(cards[i].csr_base, cards[i].csr_len);
    release_mem_region(cards[i].mem_base, cards[i].mem_len);
    printk("MM%d: Unable to remap memory region\n",i);
    return(-ENOMEM);
  }

  printk("MM%d: MEM 0x%8lx -> 0x%8lx (0x%lx)\n",i,cards[i].mem_base, cards[i].mem_remap,cards[i].mem_len);
#else
  printk("MM%d: MEM area not remapped (CONFIG_MM_MAP_MEMORY not set)\n",i);
#endif
  /* Enable the card for I/O, mem, and DMA */
  //cfg_word = PCI_COMMAND_IO | PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER;
  cfg_word = PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER;
  pci_write_config_word(dev, PCI_COMMAND, cfg_word);

  if (*((unsigned char *)(cards[i].csr_remap + MEMCTRLSTATUS_MAGIC)) != (unsigned char) 0x59) {
    printk("MM%d: Magic number invalid\n", i);
    iounmap((void *) cards[i].csr_remap);
    release_mem_region(cards[i].csr_base, cards[i].csr_len);
#ifdef CONFIG_MM_MAP_MEMORY
    iounmap((void *) cards[i].mem_remap);
    release_mem_region(cards[i].mem_base, cards[i].mem_len);
#endif
    return -1;
  }

  mem_present = (unsigned char *) cards[i].csr_remap + MEMCTRLSTATUS_MEMORY;
  if (*mem_present == MEM_128_MB)
    cards[i].rd_size = 1024*128;
  else if (*mem_present == MEM_256_MB)
    cards[i].rd_size = 1024*256;
  else if (*mem_present == MEM_512_MB)
    cards[i].rd_size = 1024*512;
  else if (*mem_present == MEM_1_GB)
    cards[i].rd_size = 1024*1024;
  else if (*mem_present == MEM_2_GB)
    cards[i].rd_size = 1024*2048;
  else
    cards[i].rd_size = 0;

  batt_status = (unsigned char *) cards[i].csr_remap + MEMCTRLSTATUS_BATTERY;
  printk("MM%d: Size %d KB, Battery 1 %s (%s), Battery 2 %s (%s)\n",
         i,
         cards[i].rd_size,
         (*batt_status & 0x01 ? "Disabled" : "Enabled"),
         (*batt_status & 0x02 ? "FAILURE" : "OK"),
         (*batt_status & 0x04 ? "Disabled" : "Enabled"),
         (*batt_status & 0x08 ? "FAILURE" : "OK"));

  pci_read_config_dword(dev, PCI_BASE_ADDRESS_1, &saved_bar);
  data = 0xffffffff;
  pci_write_config_dword(dev, PCI_BASE_ADDRESS_1, data);
  pci_read_config_dword(dev, PCI_BASE_ADDRESS_1, &data);
  pci_write_config_dword(dev, PCI_BASE_ADDRESS_1, saved_bar);
  data &= 0xfffffff0;
  data = ~data;
  data +=1;
  cards[i].win_size = data;

  /* Enable ECC */
  edc_ctrl = (unsigned char *) cards[i].csr_remap + MEMCTRLCMD_ERRCTRL;
  *edc_ctrl = EDC_STORE_CORRECT;
//  *edc_ctrl = EDC_NONE;

  /* Create wait queue */
  init_waitqueue_head(&cards[i].wait);

  if (!(cards[i].dma_buffer = (unsigned char *) __get_free_page(GFP_KERNEL))) {
    printk("Unable to allocate DMA buffer\n");
    return 1;
  }

  printk("DMA buffer allocated at %p\n",cards[i].dma_buffer);
  init_timer(&cards[i].irq_timer);
  cards[i].irq_timer.data = (void *) &cards[i];
  cards[i].irq_timer.function = mm_int_timer;

  if (request_irq(dev->irq, mm_interrupt, SA_SHIRQ, "pci-mm", &cards[i])) {
    printk("Unable to allocate IRQ\n");
    return 1;
  }

  cards[i].irq = dev->irq;
  printk("MM%d: Window size %d bytes, IRQ %d\n",i, data, cards[i].irq);

//  single_test(i);
//  dword_test(i);
//  dump_regs(i);
//  window_test(i);
 // dma_test(i);
  return 0;
}

static void dump_regs(int card)
{
  unsigned char *charAddr;
  int           idx;

  //charAddr = (unsigned char *) cards[card].csr_remap + 0x40;
  charAddr = (unsigned char *) cards[card].csr_remap;

  for (idx = 0; idx < 8; idx++) {
    printk("%p   ", charAddr);
    printk("%02x ", *(charAddr));
    printk("%02x ", *(charAddr + 1));
    printk("%02x ", *(charAddr + 2));
    printk("%02x ", *(charAddr + 3));
    printk("%02x ", *(charAddr + 4));
    printk("%02x ", *(charAddr + 5));
    printk("%02x ", *(charAddr + 6));
    printk("%02x ", *(charAddr + 7));
    printk("%02x ", *(charAddr + 8));
    printk("%02x ", *(charAddr + 9));
    printk("%02x ", *(charAddr + 10));
    printk("%02x ", *(charAddr + 11));
    printk("%02x ", *(charAddr + 12));
    printk("%02x ", *(charAddr + 13));
    printk("%02x ", *(charAddr + 14));
    printk("%02x ", *(charAddr + 15));
    charAddr += 16;
    printk("\n");
  }
}

#ifdef CONFIG_MM_MAP_MEMORY
static void dump_card(int card, int nbytes)
{
  unsigned char *p;
  int           i;

  p = (unsigned char *) cards[card].mem_remap;

  printk("MM%d: DUMP (%d)\n", card, nbytes);
  for (i = 0; i < nbytes; i++)
    printk("%x ",*p++);

  printk("\nMM%d: END DUMP\n",card);
}

static int single_test(int card)
{
  unsigned char *buffer;
  unsigned int i;

  printk("MM%d: Performing single byte access test\n", card);
  buffer = (unsigned char *) cards[card].mem_remap;
  dump_card(card, 2);

  return 0;
}

static int dword_test(int card)
{
  volatile unsigned int *buffer;
  unsigned int dword;
  unsigned int i;

  buffer = (unsigned int *) (cards[card].mem_remap);

  printk("MM%d: Performing dword access test at addr 0x%p\n", card, buffer);
  dword = 0xdeadbabe;

  *buffer = dword;
  printk("Dword at %p now = 0x%x\n", buffer, *buffer);
  printk("Dword at %p now = 0x%x\n", buffer, *buffer);
  printk("Dword at %p now = 0x%x\n", buffer, *buffer);

#if 0
  dword = *buffer;
  if (dword != (unsigned int) 0xdeadbabe) {
    printk("MM%d: Dword compare failed (0x%x)\n", card, dword);
    return 1;
  }
#endif

  return 0;
}

static int window_test(int card)
{
  unsigned char *buffer;
  unsigned int i;

  printk("MM%d: Performing memory window test\n",card);
  buffer = (unsigned char *) cards[card].mem_remap;
  printk("MM%d: Window address %p\n",card,buffer);

  dump_card(card, 16);
  for (i = 0; i < 255; i++)
    buffer[i] = 0x55;

  dump_card(card, 16);

  for (i = 0; i < 255; i++)
    if (buffer[i] != 0x55) {
      printk("MM%d: Window compare failed (offset %d)\n",card,i);
      return(1);
    }

  return 0;
}
#endif

#define NUM_ITERATIONS 1024
static int dma_test(int card)
{
  unsigned char *buffer;
  unsigned int idx, i, idx2;
  unsigned long host, local;

  printk("MM%d: Performing DMA test\n",card);
  buffer = (unsigned char *) cards[card].dma_buffer;

  while ((unsigned long) buffer & 0x7)
    buffer++;

  for (idx=0, i=0; idx < 1024; idx++) {
    buffer[idx] = i++;
    if (i == 0xff)
      i = 0;
  }

  local = 0x0;
  host = virt_to_bus((void *) buffer);

#if 0
  printk("MM%d: Send buffer VIRT %p, BUS %p\n",
         card,
         buffer,
         (void *) host);
#endif
  for (idx2 = 0; idx2 < NUM_ITERATIONS; idx2++) {
    local = 1024*idx2;
    if (do_dma(card,
               host,
               local,
               1024,
               DMA_READ_FROM_HOST) != 0) {
      printk("MM%d: DMA Write test failed\n",card);
      return 1;
    }
  }

  memset(buffer, 0, 1024);

  local = 0x00;
  host = virt_to_bus((void *) buffer);

#if 0
  printk("MM%d: Recv buffer VIRT %p, BUS %p\n",
         card,
         buffer,
         (void *) host);
#endif

  for (idx2 = 0; idx2 < NUM_ITERATIONS; idx2++) {
    local = 1024 * idx2;

    if (do_dma(card,
               host,
               local,
               1024,
               DMA_WRITE_TO_HOST) != 0) {
      printk("MM%d: DMA Read test failed\n",card);
      return 1;
    }

    for (idx=0, i=0; idx < 1024; idx++) {
      if (buffer[idx] != i++) {
        printk("MM%d: DMA compare failed (offset %d)\n",card,((1024 * idx2) + idx));
        return 1;
      }

      if (i == 0xff)
        i = 0x00;
    }
  }

  printk("MM%d: DMA test PASSED\n",card);
  return 0;
}


static void dump_dmastat(int card, unsigned int dmastat)
{
  printk("MM%d*: DMAstat - ",card);
  if (dmastat & DMASTAT_ANYERROR)
    printk("ANY ");
  if (dmastat & DMASTAT_MBEERROR)
    printk("MBE ");
  if (dmastat & DMASTAT_PERR_REP)
    printk("PERR_REP ");
  if (dmastat & DMASTAT_PERR_DET)
    printk("PERR_DET ");
  if (dmastat & DMASTAT_SERR_SIG)
    printk("SERR_SIG ");
  if (dmastat & DMASTAT_TARG_ABT)
    printk("TARG_ABT ");
  if (dmastat & DMASTAT_MAST_ABT)
    printk("MAST_ABT ");
  if (dmastat & DMASTAT_COMPLETE)
    printk("DMA_COMP ");
  printk("\n");
}

// #define CONFIG_MM_INT_DEBUG
static void mm_interrupt(int irq, void *__card, struct pt_regs *regs)
{
  struct cardinfo *card = (struct cardinfo *) __card;
  unsigned int    *dma_reg;
  unsigned int    *dma_count;

  dma_count = (unsigned int *)((unsigned long) 
            card->csr_remap + DMA_TRANSFER_SIZE);

  dma_reg = (unsigned int *)((unsigned long)
          card->csr_remap + DMA_STATUS_CTRL);

#ifdef CONFIG_MM_INT_DEBUG
  dump_dmastat(card->card_number, *dma_reg);
#endif

  if (*dma_reg & DMASTAT_ANYERROR) {
    unsigned char *err_stat;
    unsigned char stat;
    unsigned int  *err_data_log;
    unsigned int  *err_addr_log;
    unsigned char *check, *syn, *cnt;

    err_stat = (unsigned char *) card->csr_remap + 0x0E;
    err_data_log = (unsigned int *) ((unsigned long)card->csr_remap + 0x20);
    err_addr_log = (unsigned int *) ((unsigned long)card->csr_remap + 0x28);
    check = (unsigned char *) card->csr_remap + 0x2f;
    syn = (unsigned char *) card->csr_remap + 0x2e;
    cnt = (unsigned char *) card->csr_remap + 0x0d;

    dump_dmastat(card->card_number, *dma_reg);
    stat = *err_stat;
    if (stat & 0x01)
      printk("MM%d*: Memory access error detected (err count %d)\n",card->card_number, *cnt);
    if (stat & 0x02)
      printk("MM%d*: Multi-bit EDC error\n",card->card_number);

    printk("MM%d*: Fault Address 0x%x, Fault Data 0x%x\n",
           card->card_number,
           *err_addr_log,
           *err_data_log);
    printk("MM%d*: Fault Check 0x%x, Fault Syndrome 0x%x\n",
           card->card_number,
           *check,
           *syn);

    *cnt = 0;
    card->error = 1;
    wake_up(&card->wait);
  }

  if (*dma_reg & DMASTAT_MAST_ABT) {
    unsigned short cfg_status;

    printk("MM%d*: MASTER ABORT\n", card->card_number); 
    pci_read_config_word(card->dev, PCI_STATUS, &cfg_status);
    cfg_status &= ~PCI_STATUS_SIG_TARGET_ABORT;
    pci_write_config_word(card->dev, PCI_STATUS, cfg_status);
    card->error = 1;
    wake_up(&card->wait);
  } else if (*dma_count == 0) {
    card->error = 0;
    wake_up(&card->wait);
  }

  // Ack interrupt
  if (*dma_reg & DMASTAT_COMPLETE)
    *dma_reg |= DMASTAT_COMPLETE;
}

static unsigned char get_window(int card)
{
  unsigned char *win_reg;

  win_reg = (unsigned char *) (cards[card].csr_remap + WINDOWMAP_WINNUM);
  return(*win_reg);
}

static void set_window(int card, unsigned char num)
{
  unsigned char *win_reg;

  win_reg = (unsigned char *) (cards[card].csr_remap + WINDOWMAP_WINNUM);
  *win_reg = (unsigned char) num;
}

static int set_removeled(int card, unsigned char state)
{
  unsigned char *led_ctrl;
  unsigned char led_bkup;

  led_ctrl = (unsigned char *) cards[card].csr_remap + MEMCTRLCMD_LEDCTRL;
  led_bkup = *led_ctrl;
  led_bkup = led_bkup & 0xF3;
  led_bkup|= (state << 2);
  *led_ctrl = led_bkup;

  return 0;
}

static void mm_int_timer(unsigned long data)
{
  struct cardinfo *card = (struct cardinfo *) data;

#ifdef CONFIG_MM_INT_DEBUG
  printk("MM%d: IRQ timeout\n", card->card_number);
#endif
  card->error = 2;
  wake_up(&card->wait);
}

#ifdef MODULE
module_init(mm_init);
module_exit(mm_cleanup);
#endif
